////////////////////////////////////////////////////////////////////////
/*

     File: DataSavePanel.java
   Author: Peter Hollemans
     Date: 2004/05/04

  CoastWatch Software Library and Utilities
  Copyright (c) 2004 National Oceanic and Atmospheric Administration
  All rights reserved.

  Developed by: CoastWatch / OceanWatch
                Center for Satellite Applications and Research
                http://coastwatch.noaa.gov

  For conditions of distribution and use, see the accompanying
  license.txt file.

*/
////////////////////////////////////////////////////////////////////////

// Package
// -------
package noaa.coastwatch.gui.save;

// Imports
// -------
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.Iterator;
import java.util.List;
import javax.swing.JOptionPane;
import javax.swing.ProgressMonitor;
import javax.swing.Timer;
import noaa.coastwatch.gui.GUIServices;
import noaa.coastwatch.gui.save.ArcSavePanel;
import noaa.coastwatch.gui.save.BinarySavePanel;
import noaa.coastwatch.gui.save.CWHDFSavePanel;
import noaa.coastwatch.gui.save.SavePanel;
import noaa.coastwatch.gui.save.SubsetOptionPanel;
import noaa.coastwatch.gui.save.TextSavePanel;
import noaa.coastwatch.gui.save.VariableOptionPanel;
import noaa.coastwatch.io.EarthDataReader;
import noaa.coastwatch.io.EarthDataWriter;
import noaa.coastwatch.render.ColorComposite;
import noaa.coastwatch.render.ColorEnhancement;
import noaa.coastwatch.render.EarthDataView;
import noaa.coastwatch.util.DataLocation;
import noaa.coastwatch.util.EarthDataInfo;
import noaa.coastwatch.util.EarthLocation;
import noaa.coastwatch.util.Grid;
import noaa.coastwatch.util.trans.Datum;
import noaa.coastwatch.util.trans.EarthTransform;

/** 
 * The <code>DataSavePanel</code> class is the abstract parent of all
 * data save panels, which allow the user to export earth data to
 * various file formats.  To export data, create a new panel with the
 * <code>create()</code> method, then call the <code>write()</code>
 * method.  Generally, the save panel is enclosed in a dialog with OK
 * and Cancel options.
 *
 * @author Peter Hollemans
 * @since 3.1.7
 */
public abstract class DataSavePanel
  extends SavePanel {

  // Variables 
  // ---------

  /** The earth data reader to use for data. */
  protected EarthDataReader reader;

  /** The list of variables for export. */
  protected List variableList;

  /** The current earth view. */
  protected EarthDataView view;

  /** The subset panel used for creating data subsets. */
  protected SubsetOptionPanel subsetPanel;

  /** The variable option panel. */
  protected VariableOptionPanel variablePanel;

  ////////////////////////////////////////////////////////////

  /** Sets the list of selected variable names. */
  public void setVariables (List nameList) { 

    variablePanel.setVariables (nameList);

  } // setVariables

  ////////////////////////////////////////////////////////////

  /** 
   * Returns true if this class requires single variable selection in
   * the variable list.  By default this method returns false unless
   * overridden in the child class.
   */
  protected boolean isSingleVariable() { return (false); }

  ////////////////////////////////////////////////////////////

  /* 
   * Creates a new panel appropriate for the specified format. 
   *
   * @param reader the reader to use as a source of data.
   * @param variableList the list of available variables to export.
   * @param view the current data view.
   * @param format the file format.
   */
  public static DataSavePanel create (
    EarthDataReader reader,
    List variableList,
    EarthDataView view,
    String format
  ) {

    // Create new panel
    // ----------------
    DataSavePanel panel;
    if (format.equals ("hdf"))
      panel = new CWHDFSavePanel (reader, variableList, view);
    else if (format.equals ("nc"))
      panel = new CFNCSavePanel (reader, variableList, view);
    else if (format.equals ("nc4"))
      panel = new CFNC4SavePanel (reader, variableList, view);
    else if (format.equals ("raw"))
      panel = new BinarySavePanel (reader, variableList, view);
    else if (format.equals ("txt"))
      panel = new TextSavePanel (reader, variableList, view);
    else if (format.equals ("flt"))
      panel = new ArcSavePanel (reader, variableList, view);
    else 
      throw new IllegalArgumentException ("Unsupported format: " + format);

    return (panel);

  } // create

  ////////////////////////////////////////////////////////////

  /* 
   * Creates a new image save panel.
   *
   * @param reader the reader to use as a source of data.
   * @param variableList the list of available variables to export.
   * @param view the current data view.
   */
  protected DataSavePanel (
    EarthDataReader reader,
    List variableList,
    EarthDataView view
  ) {

    // Initialize
    // ----------
    this.reader = reader;
    this.variableList = variableList;
    this.view = view;

    // Create subset option panel
    // --------------------------
    subsetPanel = createSubsetPanel();
    variablePanel = new VariableOptionPanel (variableList, isSingleVariable());

  } // DataSavePanel constructor

  ////////////////////////////////////////////////////////////

  /** Checks the panel contents. */
  public void check () {

    // Check subset
    // ------------
    if (subsetPanel.getSubsetType() != SubsetOptionPanel.SUBSET_ALL) {
      int[] start = new int[2];
      int[] dims = new int[2];
      getPanelSubset (start, dims);
      Grid grid = getDataGrid();
      if (!grid.checkSubset (start, dims))
        throw new RuntimeException ("Invalid subset specified");
    } // if

    // Check list of variables
    // -----------------------
    List variableList = variablePanel.getVariables();
    if (variableList.size() == 0) {
      throw new RuntimeException ("No variables were selected for export");
    } // if

  } // check

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the subset information specified by the subset option
   * panel. 
   *
   * @param upperLeft the upper-left data coordinate as <code>[row,
   * column]</code> (modified).
   * @param dimensions the data dimensions as <code>[rows,
   * columns]</code> (modified).
   */
  protected void getPanelSubset (
    int[] upperLeft,
    int[] dimensions
  ) {

    int type = subsetPanel.getSubsetType();
    switch (type) {

    // Get current view subset
    // -----------------------      
    case SubsetOptionPanel.SUBSET_CURRENT:
      getViewSubset (upperLeft, dimensions, null);
      break;

    // Get image coordinate-based subset
    // ---------------------------------
    case SubsetOptionPanel.SUBSET_IMAGE:
      subsetPanel.getImageCoords (upperLeft, dimensions);
      break;

    // Get geographic coordinate-based subset
    // --------------------------------------
    case SubsetOptionPanel.SUBSET_GEOGRAPHIC:

      // Get bounds
      // ----------
      EarthTransform trans = reader.getInfo().getTransform();
      Datum datum = trans.getDatum();
      EarthLocation[] bounds = new EarthLocation[2];
      bounds[0] = new EarthLocation (datum);
      bounds[1] = new EarthLocation (datum);
      subsetPanel.getGeographicCoords (bounds[0], bounds[1]);

      // Transform bounds to corners
      // ---------------------------
      DataLocation[] corners = new DataLocation[2];
      corners[0] = trans.transform (bounds[0]);
      corners[1] = trans.transform (bounds[1]);
      if (!corners[0].isValid() || !corners[1].isValid())
        throw new RuntimeException ("Invalid geographic corners specified");

      // Get truncated corners
      // ---------------------
      getTruncated (corners, upperLeft, dimensions);

      break;
      
    default: 
      throw new IllegalStateException ("Unknown subset type specified");

    } // switch

  } // getPanelSubset

  ////////////////////////////////////////////////////////////

  /**
   * Creates an information object based on the reader, taking into
   * account any subset specified.
   *
   * @return the new information object.
   */
  protected EarthDataInfo createInfo () {

    EarthDataInfo info = reader.getInfo();
    EarthDataInfo newInfo = (EarthDataInfo) info.clone();
    if (subsetPanel.getSubsetType() != SubsetOptionPanel.SUBSET_ALL) {
      int[] upperLeft = new int[2];
      int[] dimensions = new int[2];
      getPanelSubset (upperLeft, dimensions);
      DataLocation upperLeftLoc = 
        new DataLocation (upperLeft[0], upperLeft[1]);
      EarthTransform newTrans = 
        info.getTransform().getSubset (upperLeftLoc, dimensions);
      newInfo.setTransform (newTrans);
    } // if

    return (newInfo);

  } // createInfo

  ////////////////////////////////////////////////////////////

  /** Gets the data grid in the view. */
  private Grid getDataGrid () {

    Grid grid;
    if (view instanceof ColorEnhancement) 
      grid = ((ColorEnhancement) view).getGrid();
    else if (view instanceof ColorComposite)
      grid = ((ColorComposite) view).getGrids()[0];
    else 
      throw new IllegalStateException ("Unknown earth data view");

    return (grid);

  } // getDataGrid

  ////////////////////////////////////////////////////////////

  /** Gets the full data dimensions of the grid. */
  private int[] getDataDimensions () {

    return (getDataGrid().getDimensions());

  } // getDataDimensions

  ////////////////////////////////////////////////////////////
  
  /** 
   * Gets the integer data location coordinates of the specified data
   * locations.
   *
   * @param corners the data location corners as <code>[upper-left,
   * lower-right]</code> (modified).
   * @param lowerRightLoc the lower-right data location.
   * @param upperLeft the upper-left data coordinate as <code>[row,
   * column]</code> (modified).
   * @param dimensions the data dimensions as <code>[rows,
   * columns]</code> (modified).
   */
  private void getTruncated (
    DataLocation[] corners,
    int[] upperLeft,
    int[] dimensions
  ) {

    int[] dims = getDataDimensions();
    corners[0] = corners[0].truncate (dims).round();
    corners[1] = corners[1].truncate (dims).round();
    upperLeft[0] = (int) corners[0].get (0);
    upperLeft[1] = (int) corners[0].get (1);
    dimensions[0] = ((int) corners[1].get(0)) - ((int) corners[0].get(0)) + 1;
    dimensions[1] = ((int) corners[1].get(1)) - ((int) corners[0].get(1)) + 1;

  } // getTruncated

  ////////////////////////////////////////////////////////////

  /** 
   * Gets the subset information specified by the view.
   *
   * @param upperLeft the upper-left data coordinate as <code>[row,
   * column]</code> (modified).
   * @param dimensions the data dimensions as <code>[rows,
   * columns]</code> (modified).
   * @param bounds the data bounds as
   * <code>[upper-left, lower-right]</code> (modified) or null.
   */
  private void getViewSubset (
    int[] upperLeft,
    int[] dimensions,
    EarthLocation[] bounds
  ) {

    // Get truncated corners
    // ---------------------
    DataLocation[] corners = view.getBounds();
    getTruncated (corners, upperLeft, dimensions);

    // Get bounds
    // ----------
    if (bounds != null) {
      EarthTransform trans = reader.getInfo().getTransform();
      bounds[0] = trans.transform (corners[0]);
      bounds[1] = trans.transform (corners[1]);
    } // if

  } // getViewSubset

  ////////////////////////////////////////////////////////////

  /** 
   * Creates a subset option panel whose parameters match the
   * view.
   */
  private SubsetOptionPanel createSubsetPanel () {

    // Get view subset
    // ---------------
    int[] upperLeft = new int[2];
    int[] dimensions = new int[2];
    EarthLocation[] bounds = new EarthLocation[2];
    getViewSubset (upperLeft, dimensions, bounds);

    return (new SubsetOptionPanel (upperLeft, dimensions, bounds));

  } // createSubsetPanel

  ////////////////////////////////////////////////////////////

  /** 
   * Writes the list of variables from the variable option panel to
   * the writer.  The subset option panel is used to subset the
   * variable data if needed.  The writer is closed after the final
   * variable data is written.
   *
   * @param writer the writer used to export data.
   *
   * @throws IOException if an error occurred writing the data.
   */
  protected void write (
    final EarthDataWriter writer
  ) throws IOException {
  
    // Get subset parameters
    // ---------------------
    boolean isSubset = 
      (subsetPanel.getSubsetType() != SubsetOptionPanel.SUBSET_ALL);
    int[] start = new int[2];
    int[] dims = new int[2];
    if (isSubset) getPanelSubset (start, dims);

    // Add each variable to writer
    // ---------------------------
    List variableList = variablePanel.getVariables();
    for (Iterator iter = variableList.iterator(); iter.hasNext(); ) {
      String variableName = (String) iter.next();
      Grid grid = (Grid) reader.getVariable (variableName);
      if (isSubset) grid = grid.getSubset (start, dims);
      writer.addVariable (grid);
    } // for

    // Create progress monitor
    // -----------------------
    final File file = new File (writer.getDestination());
    final ProgressMonitor monitor =  new ProgressMonitor (this, 
      "Saving data to " + file.getName(), " ", 0, writer.getProgressLength());
    monitor.setMillisToPopup (1000);

    // Create progress update task
    // ---------------------------
    ActionListener task = new ActionListener() {
        public void actionPerformed (ActionEvent event) {
          monitor.setProgress (writer.getProgress());
          String name = writer.getProgressVariable();
          if (name != null) monitor.setNote ("Writing " + name);
          if (monitor.isCanceled()) writer.cancel();
        } // actionPerformed
      };
    final Timer timer = new Timer (500, task);

    // Perform data write
    // ------------------
    Thread writeThread = new Thread () {
        public void run () {

          // Attempt write and close
          // -----------------------
          try { 
            writer.flush();
            writer.close(); 
          } // try

          // Show error message
          // ------------------
          catch (IOException e) {
            final String errorMessage = 
              "An error occurred writing the file:\n" + e.toString();
            GUIServices.invokeAndWait (new Runnable() {
                public void run() {
                  JOptionPane.showMessageDialog (DataSavePanel.this,
                    errorMessage, "Error", JOptionPane.ERROR_MESSAGE);
                } // run
              });
          } // catch

          // Stop timer and close monitor dialog
          // -----------------------------------
          finally {
            timer.stop();
            if (monitor.isCanceled()) {
              if (file.exists()) file.delete();
            } // if
            else {
              GUIServices.invokeAndWait (new Runnable() {
                  public void run() {
                    monitor.close();
                  } // run
                });
            } // else
          } // finally

        } // run
      };
    writeThread.start();
    timer.start();

    // TODO: A problem can occur if the write thread is writing data
    // while another thread is attempting to read data from the same
    // file.  Partly, this is because the progress dialog does not
    // seem to be modal and allows operations in other windows.
    // Another reason is that file I/O is not always inside a
    // synchronized statement when it should be.  For example, if two
    // threads are both using the same JNI library.

  } // write

  ////////////////////////////////////////////////////////////

} // DataSavePanel class

////////////////////////////////////////////////////////////////////////
